/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Copyright (C) 2004 Andy Wingo <wingo@pobox.com>
 * Copyright (C) 2003 Imendio HB 
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
 
#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-mime.h>
#include <libgnomevfs/gnome-vfs-module.h>


#define DEBUG_PRINT

#ifdef DEBUG_PRINT
#define DEBUG(string,args...) g_print ("floppy debug: " string "\n", ##args);
#else
#define DEBUG(string,args...)
#endif


/* it's really const, but argv doesn't like that */
static gchar*
uri_to_path (GnomeVFSURI *uri) 
{
	/* strip leading slashes */
	gchar *path = uri->text;
	while (*path == '/') path++;
	return path;
}

static GnomeVFSResult
mtools_do (gchar **argv, gchar ***lines_ret, GnomeVFSContext *context)
{
	gboolean success = FALSE;
	gchar *out = NULL, *err = NULL;
	gint ecode = -1;
	GError *error = NULL;
	GnomeVFSResult res;
	static GStaticMutex mtools_lock = G_STATIC_MUTEX_INIT;

	DEBUG ("mtools_do: %s", argv[0]);

	if (lines_ret)
		*lines_ret = NULL;

	/* mtools access will fail if the floppy is busy; serialize access with
	   this lock so nautilus doesn't complain. */
	g_static_mutex_lock (&mtools_lock);

	if (gnome_vfs_context_check_cancellation (context)) {
		g_static_mutex_unlock (&mtools_lock);
		DEBUG ("%s cancelled", argv[0]);
		g_free (argv);
		return GNOME_VFS_ERROR_CANCELLED;
	}
	
	success = g_spawn_sync (g_get_tmp_dir (),
				argv,  /* argv */
				NULL,  /* environment */
				0,     /* flags */
				NULL,  /* child setup proc */
				NULL,  /* child setup data */
				&out,  /* stdout output */
				&err,  /* stderr output */
				&ecode,/* exit code */
				&error);
	g_static_mutex_unlock (&mtools_lock);
	g_free (argv);

	/* should catch :
	   Cannot initialize
	   [corruption error]
	   file not found
	   mtools not there!
	*/

	if (!success) {
		g_warning ("Call to %s failed: %s", argv[0],
			   error ? error->message : "(unknown error!)");
		if (error)
			g_error_free (error);

		return GNOME_VFS_ERROR_NOT_SUPPORTED;
	} else if (ecode) {
		if (strstr (err, "not found")) {
			DEBUG ("File not found");
			res = GNOME_VFS_ERROR_NOT_FOUND;
		} else {
			DEBUG ("Unknown error: %s", err);
			res = GNOME_VFS_ERROR_GENERIC;
		}
		
		g_free (out);
		g_free (err);
	} else {
		if (lines_ret)
			*lines_ret = g_strsplit (out, "\n", 0);
		g_free (out);
		g_free (err);
		DEBUG ("Successful");
		res = GNOME_VFS_OK;
	}

	return res;
}

static GnomeVFSResult
mdir_advance_current (gchar **lines, gint *current)
{
	gchar *line;
	gint i = *current;
	
	line = lines[i];
	if (strstr (line, "not found")) {
		g_strfreev (lines);
		DEBUG ("Not found");
		return GNOME_VFS_ERROR_NOT_FOUND;
	}

	for (line = lines[i];
	     line && (strncmp (line, "Directory for ", 14) != 0);
	     line = lines[++i]);

	if (line) {
		line = lines[++i];
		if (!line || (strlen (line) != 0)) {
			g_warning ("failed to parse mdir output: %s",
				   line ? line : "(null)");
			g_strfreev (lines);
			return GNOME_VFS_ERROR_INTERNAL;
		}
		i++;
		*current = i;
		return GNOME_VFS_OK;
	} else {
		g_warning ("failed to parse mdir output");
		g_strfreev (lines);
		return GNOME_VFS_ERROR_INTERNAL;
	}
}

static GnomeVFSResult
parse_mdir_line (const gchar *line, GnomeVFSFileInfo *file_info)
{
	gint len = line ? strlen (line) : 0;

	if (!line                               /* unexpected end of argv */
	    || (len == 0)                       /* "", happens at the very end */
	    || (strcmp (line, "No files") == 0) /* if the root dir is empty */
	    || line[0] == ' ') {                /* end of dir list */
		DEBUG ("EOF reading dir");
		return GNOME_VFS_ERROR_EOF;
	}

	/* layout of line:
	   0    5    10   15   20   25   30   35   40   45   50   55
	   8CHARNAM EXT SIZE<DIR> YYYY-MM-DD  HH:MM  Long name till EOL */
	
	if (line[0] == '.') {
		if (line[1] == '.') {
			file_info->name = g_strdup ("..");
		} else {
			file_info->name = g_strdup (".");
		}
	} else {
		if (len < 43) {
			g_warning ("couldn't parse mdir line: %s", line);
			return GNOME_VFS_ERROR_INTERNAL;
		}
		file_info->name = g_strdup (&line[42]);
	}

	if (line[13] == '<') {
		/* it's a DIR (showing <DIR> in the results) */
		file_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_TYPE;
		file_info->mime_type = g_strdup ("x-directory/normal");
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
	} else {
		file_info->type = GNOME_VFS_FILE_TYPE_REGULAR;
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_TYPE;
		file_info->mime_type =
			g_strdup (gnome_vfs_get_file_mime_type_fast (file_info->name,
								     NULL));
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
		file_info->size = atol (&line[13]);
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_SIZE;
		file_info->permissions =
			GNOME_VFS_PERM_USER_READ | GNOME_VFS_PERM_USER_WRITE
			| GNOME_VFS_PERM_GROUP_READ | GNOME_VFS_PERM_GROUP_WRITE
			| GNOME_VFS_PERM_OTHER_READ | GNOME_VFS_PERM_OTHER_WRITE
			| GNOME_VFS_PERM_ACCESS_READABLE | GNOME_VFS_PERM_ACCESS_WRITABLE;
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_ACCESS;
		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS;
	}

	return GNOME_VFS_OK;
}


static GHashTable *floppycache;
G_LOCK_DEFINE_STATIC (lock);

typedef struct {
	gint current;
	gchar **entries;
} DirCache;

typedef struct {
	int ref_count;
	gchar *path;	
	gchar *tmp_file;
} FloppyCache;

typedef struct {
	FloppyCache *cache;
	int fd;
	GnomeVFSOpenMode mode;
} FloppyFile;

static GnomeVFSResult
floppy_cache_create (FloppyCache **cache_ret, const gchar *path)
{
	int fd;
	FloppyCache *cache;
	gchar **argv;
	gchar *tmp_file;
	GnomeVFSResult res;
	
	fd = g_file_open_tmp ("gnome-mtoolsXXXXXX",
			      &tmp_file, NULL);
	if (fd < 0) {
		G_UNLOCK (lock);
		g_warning ("failed to create the tmp file");
		*cache_ret = NULL;
		return GNOME_VFS_ERROR_ACCESS_DENIED;
	}
	close (fd); fd = -1;

	argv = g_new0 (gchar*, 5);
	argv[0] = "/usr/bin/mcopy";
	argv[1] = "-n";
	argv[2] = (gchar*)path; /* it's ok! */
	argv[3] = tmp_file;
	argv[4] = NULL;
	res = mtools_do (argv, NULL, NULL);
			
	if (res != GNOME_VFS_OK) {
		*cache_ret = NULL;
		return res;
	}

	cache = g_new0 (FloppyCache, 1);
	cache->path = g_strdup (path);
	cache->ref_count = 1;
	cache->tmp_file = tmp_file;
	*cache_ret = cache;
	return GNOME_VFS_OK;
}

static GnomeVFSResult
floppy_cache_sync (FloppyCache *cache)
{
	gchar **argv;
	
	argv = g_new0 (gchar*, 4);
	argv[0] = "/usr/bin/mcopy";
	argv[1] = cache->tmp_file;
	argv[2] = cache->path;
	argv[3] = NULL;
	return mtools_do (argv, NULL, NULL);
}

static void
floppy_cache_destroy (FloppyCache *cache)
{
	g_assert (cache->ref_count == 0);
	g_free (cache->path);
	g_free (cache->tmp_file);
	g_free (cache);
}
	
static GnomeVFSResult
floppy_file_open (FloppyFile **file_ret, const gchar *path,
		  GnomeVFSOpenMode mode)
{
	GnomeVFSResult res;
	FloppyFile *file;
	FloppyCache *cache;
	int fd = -1;
	
	*file_ret = NULL;

	G_LOCK (lock);

	cache = g_hash_table_lookup (floppycache, path);
	if (!cache) {
		res = floppy_cache_create (&cache, path);
		if (res != GNOME_VFS_OK) {
			G_UNLOCK (lock);
			return res;
		}
	} else {
		cache->ref_count++;
	}
	
	fd = open (cache->tmp_file,
		   mode & GNOME_VFS_OPEN_WRITE ? O_WRONLY : O_RDONLY);
	if (fd < 0) {
		g_warning ("failed to open the cache file!");
		if (--cache->ref_count == 0)
			floppy_cache_destroy (cache);
		G_UNLOCK (lock);
		return GNOME_VFS_ERROR_ACCESS_DENIED;
	}

	G_UNLOCK (lock);
	
	file = g_new0 (FloppyFile, 1);
	file->cache = cache;
	file->mode = mode;
	file->fd = fd;

	*file_ret = file;
	return GNOME_VFS_OK;
}

static GnomeVFSResult
floppy_file_close (FloppyFile *file)
{
	GnomeVFSResult res = GNOME_VFS_OK;
	FloppyCache *cache = file->cache;
	
	close (file->fd);

	if (file->mode & GNOME_VFS_OPEN_WRITE)
		res = floppy_cache_sync (cache);
	
	G_LOCK (lock);

	if (--cache->ref_count == 0) {
		g_hash_table_remove (floppycache, cache->path);
		floppy_cache_destroy (cache);
	}

	G_UNLOCK (lock);

	g_free (file);

	return res;
}

/* Implementation of GnomeVFS module functions. */

static GnomeVFSResult
do_open (GnomeVFSMethod        *method,
	 GnomeVFSMethodHandle **method_handle,
	 GnomeVFSURI           *uri,
	 GnomeVFSOpenMode       mode,
	 GnomeVFSContext       *context)
{
	FloppyFile *file;
	const gchar *path = uri_to_path (uri);
	GnomeVFSResult res;
	
	/* We don't support random mode. */
	if (mode & GNOME_VFS_OPEN_RANDOM) {
		DEBUG ("Random mode not supported");
		return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
	}

	res = floppy_file_open (&file, path, mode);
	if (res != GNOME_VFS_OK) {
		*method_handle = NULL;
		return res;
	}
	
	*method_handle = (GnomeVFSMethodHandle *) file;	
		
	return GNOME_VFS_OK;
}

static GnomeVFSResult
do_create (GnomeVFSMethod        *method,
	   GnomeVFSMethodHandle **method_handle,
	   GnomeVFSURI           *uri,
	   GnomeVFSOpenMode       mode,
	   gboolean               exclusive,
	   guint                  perm,
	   GnomeVFSContext       *context)
{
	/* We cheat here and don't take perm or exclusive in consideration. */
	return do_open (method, method_handle, uri, mode, context);
}

static GnomeVFSResult
do_read (GnomeVFSMethod       *method,
	 GnomeVFSMethodHandle *method_handle,
	 gpointer              buffer,
	 GnomeVFSFileSize      bytes,
	 GnomeVFSFileSize     *bytes_read,
	 GnomeVFSContext      *context)
{
	FloppyFile *file = (FloppyFile*)method_handle;
	gint read_val;

        do {
                read_val = read (file->fd, buffer, bytes);
        } while (read_val == -1
                 && errno == EINTR
                 && ! gnome_vfs_context_check_cancellation (context));
 
        if (read_val == -1) {
                *bytes_read = 0;
                return gnome_vfs_result_from_errno ();
        } else {
                *bytes_read = read_val;
		DEBUG ("read %d bytes", read_val);
 
                /* Getting 0 from read() means EOF! */
                if (read_val == 0) {
			DEBUG ("EOF while reading");
                        return GNOME_VFS_ERROR_EOF;
                }
        }
	return GNOME_VFS_OK;
}

static GnomeVFSResult
do_write (GnomeVFSMethod       *method,
	  GnomeVFSMethodHandle *method_handle,
	  gconstpointer         buffer,
	  GnomeVFSFileSize      bytes,
	  GnomeVFSFileSize     *bytes_written,
	  GnomeVFSContext      *context)
{
	FloppyFile *file = (FloppyFile*)method_handle;
        gint write_val;
 
        do
                write_val = write (file->fd, buffer, bytes);
        while (write_val == -1
               && errno == EINTR
               && ! gnome_vfs_context_check_cancellation (context));
	
        if (write_val == -1) {
                *bytes_written = 0;
                return gnome_vfs_result_from_errno ();
        } else {
                *bytes_written = write_val;
                return GNOME_VFS_OK;
        }
}

static GnomeVFSResult
do_close (GnomeVFSMethod       *method,
	  GnomeVFSMethodHandle *method_handle,
	  GnomeVFSContext      *context)
{
	FloppyFile *file = (FloppyFile*)method_handle;

	return floppy_file_close (file);
}

static GnomeVFSResult
do_open_directory (GnomeVFSMethod           *method,
		   GnomeVFSMethodHandle    **method_handle,
		   GnomeVFSURI              *uri,
		   GnomeVFSFileInfoOptions   options,
		   GnomeVFSContext          *context)
{
	DirCache *handle;
	gchar **argv, **lines = NULL;
	gint i = 0;
	GnomeVFSResult res;
	
	argv = g_new0 (gchar*, 3);
	argv[0] = "/usr/bin/mdir";
	argv[1] = uri_to_path (uri);
	argv[2] = NULL;
	res = mtools_do (argv, &lines, context);
	if (res != GNOME_VFS_OK)
		return res;
	
	res = mdir_advance_current (lines, &i);
	if (res != GNOME_VFS_OK)
		return res;

	handle = g_new0 (DirCache, 1);
	handle->entries = lines;
	handle->current = i;
	
	*method_handle = (GnomeVFSMethodHandle *) handle;

	return GNOME_VFS_OK;
}

static GnomeVFSResult
do_read_directory (GnomeVFSMethod       *method,
		   GnomeVFSMethodHandle *method_handle,
		   GnomeVFSFileInfo     *file_info,
		   GnomeVFSContext      *context)
{
	DirCache *handle = (DirCache *) method_handle;
	GnomeVFSResult res;

	if (handle->current < 0) {
		DEBUG ("EOF while reading dir (again)");
		return GNOME_VFS_ERROR_EOF;
	}

	res = parse_mdir_line (handle->entries[handle->current++],
			       file_info);

	if (res != GNOME_VFS_OK) {
		handle->current = -1;
	}

	return res;
}

static GnomeVFSResult
do_close_directory (GnomeVFSMethod       *method,
		    GnomeVFSMethodHandle *method_handle,
		    GnomeVFSContext      *context)
{
	DirCache *handle = (DirCache *) method_handle;

	g_strfreev (handle->entries);
	g_free (handle);

	return GNOME_VFS_OK;
}

static GnomeVFSResult
do_get_file_info (GnomeVFSMethod          *method,
		  GnomeVFSURI             *uri,
		  GnomeVFSFileInfo        *file_info,
		  GnomeVFSFileInfoOptions  options,
		  GnomeVFSContext         *context)
{
	gchar **argv;
	gchar **lines;
	gint i;
	GnomeVFSResult res;
	
	argv = g_new0 (gchar*, 4);
	argv[0] = "/usr/bin/mdir";
	argv[1] = "-f";
	argv[2] = uri_to_path (uri);
	argv[3] = NULL;
	res = mtools_do (argv, &lines, context);
	if (res != GNOME_VFS_OK)
		return res;
	
	i = 0;
	res = mdir_advance_current (lines, &i);
	if (res != GNOME_VFS_OK)
		return res;
	
	res = parse_mdir_line (lines[i], file_info);
	
	g_free (file_info->name);
	file_info->name = g_strdup (g_basename (uri_to_path (uri)));

	g_strfreev (lines);
	
	return res;
}

static gboolean
do_is_local (GnomeVFSMethod    *method,
	     const GnomeVFSURI *uri)
{
	return TRUE; /* otherwise the gtk file chooser barfs at us */
}

static GnomeVFSResult
do_make_directory (GnomeVFSMethod  *method,
		   GnomeVFSURI     *uri,
		   guint            perm,
		   GnomeVFSContext *context)
{
	gchar **argv;
	
	argv = g_new0 (gchar*, 5);
	argv[0] = "/usr/bin/mmd";
	argv[1] = "-D";
	argv[2] = "s"; /* skip if it failed */
	argv[3] = uri_to_path (uri);
	argv[4] = NULL;
	return mtools_do (argv, NULL, context);
}

static GnomeVFSResult
do_remove_directory (GnomeVFSMethod  *method,
		     GnomeVFSURI     *uri,
		     GnomeVFSContext *context)
{
	gchar **argv;
	
	argv = g_new0 (gchar*, 5);
	argv[0] = "/usr/bin/mrd";
	argv[1] = "-D";
	argv[2] = "s"; /* skip if it failed */
	argv[3] = uri_to_path (uri);
	argv[4] = NULL;
	return mtools_do (argv, NULL, context);
}

static GnomeVFSResult
do_unlink (GnomeVFSMethod  *method,
	   GnomeVFSURI     *uri,
	   GnomeVFSContext *context)
{
	gchar **argv;
	
	/* big fixme: deal with readonly files */
	argv = g_new0 (gchar*, 3);
	argv[0] = "/usr/bin/mdel";
	argv[1] = uri_to_path (uri);
	argv[2] = NULL;
	return mtools_do (argv, NULL, context);
}

static GnomeVFSMethod method = {
	sizeof (GnomeVFSMethod),
	
        do_open,                /* open */
        do_create,              /* create */
        do_close,               /* close */
        do_read,                /* read */
        do_write,               /* write */
        NULL,                   /* seek */
        NULL,                   /* tell */
        NULL,                   /* truncate_handle */ 
        do_open_directory,      /* do_open_directory */
	do_close_directory,     /* do_close_directory */
        do_read_directory,      /* do_read_directory */     
        do_get_file_info,       /* do_get_file_info */
        NULL,                   /* get_file_info_from_handle */
        do_is_local,            /* is_local */
        do_make_directory,      /* make_directory */
        do_remove_directory,    /* remove_directory */
        NULL,                   /* move */
        do_unlink,              /* unlink */
        NULL,                   /* check_same_fs */
        NULL,                   /* set_file_info */
        NULL,                   /* truncate */
        NULL,                   /* find_directory */
        NULL,                   /* create_symbolic_link */
	NULL,                   /* monitor_add */
	NULL,                   /* monitor_cancel */
	NULL                    /* file_control */
};

GnomeVFSMethod *
vfs_module_init (const char *method_name, const char *args)
{
	if (strcmp (method_name, "floppy") == 0) {
		floppycache = g_hash_table_new (g_str_hash, g_str_equal);
		return &method;
	}
	
	return NULL;
}
 
void
vfs_module_shutdown (GnomeVFSMethod* method)
{
	G_LOCK (lock);
	if (g_hash_table_size (floppycache) != 0)
		g_warning ("YIKES! Open files during shutdown!");
	else
		g_hash_table_destroy (floppycache);
	G_UNLOCK (lock);	
}

 
